@@ -16,7 +16,7 @@ use warnings;
 use lib '/usr/share/apt-cacher/lib';
 
 use sigtrap qw(die normal-signals);
-use Cwd;
+use Cwd ();
 use Fcntl qw/:DEFAULT :flock F_SETFD/;
 use Getopt::Long qw(:config no_ignore_case);
 use Digest::SHA;
@@ -122,7 +122,7 @@ load_checksum(); # Will disable checksum if BerkeleyDB not available
 if ($nice_mode) {
     printmsg("Nice mode\n");
     setpriority 0, 0, 20;
-    use IO_Prio qw(:all);
+    use Linux::IO_Prio qw(:all);
     ioprio_set(IOPRIO_WHO_PROCESS, $$,IOPRIO_PRIO_VALUE(IOPRIO_CLASS_IDLE, 0)) == 0 ||
       printmsg("ioprio_set failed: $!\n");
 }
@@ -134,6 +134,26 @@ sub printmsg {
     return $ret;
 }
 
+sub open_lock {
+    my ($file) = @_;
+
+    #  Lock header LOCK_EX first to block apt-cacher from working on it
+    open(my $hfh, '+<', "../headers/$file") || do {
+	warn ("Error: cannot open ../headers/$file for locking: $!\n");
+	return;
+    };
+    _flock($hfh, LOCK_EX) || die "Lock ../headers/$file failed: $!";
+
+    # Read lock file LOCK_SH
+    open(my $cfh, '+<', $file) || do {
+	warn ("Error: cannot open $file for locking: $!\n");
+	return;
+    };
+    _flock($cfh, LOCK_SH) || die "Lock $file failed: $!";
+
+    return ($cfh, $hfh);
+}
+
 sub get {
     my ($file, $use_url) = @_;
     my $path_info;
@@ -189,7 +209,7 @@ sub pdiff {
     }
 #    printmsg "Searching $release for $diffindex_patt and $name_patt\n";
     open(my $rfh, '<', $release) || die "Unable to open $release: $!";
-    flock($rfh, LOCK_SH);
+    _flock($rfh, LOCK_SH);
 
     my ($diffindex_sha1, $name_sha1, $name_size);
     while (<$rfh>) {
@@ -204,7 +224,7 @@ sub pdiff {
 	}
 	last if ($name_sha1 && $diffindex_sha1);
     }
-    flock($rfh, LOCK_UN);
+    _flock($rfh, LOCK_UN);
     close($rfh);
     if (!$name_sha1 || !$name_size || !$diffindex_sha1) {
         warn "SHA1s for $name_patt and/or $diffindex_patt not found in $release, aborting patch\n";
@@ -214,16 +234,17 @@ sub pdiff {
     my $sha1 = Digest::SHA->new(1); # SHA1
     my $digest;
 
+    (my ($cfh, $hfh) = open_lock($name)) == 2 || do {
+	warn "Failed to open filehandles for $name, aborting patch\n";
+	return;
+    };
+
     # Check size first
-    if (-s $name == $name_size) {
+    if (-s $cfh == $name_size) {
 	printmsg ("$name matches size in $release, going on to check SHA1..\n");
 	
 	# Check SHA1 only if size correct
-	open(my $nfh, '<', $name)|| die "Unable to open file $name for locking: $!";
-	flock($nfh, LOCK_SH);	
-	$digest = $sha1->addfile($nfh)->hexdigest;
-	flock($nfh, LOCK_UN);
-	close($nfh,);
+	$digest = $sha1->addfile($cfh)->hexdigest;
 	if ($digest eq $name_sha1) {
 	    printmsg "$name already matches SHA1 in $release: patching not required\n";
 	    return 1 # success
@@ -236,9 +257,6 @@ sub pdiff {
 	printmsg ("$name size not latest, proceeding with patch\n");
     }
 
-    open(my $lck, '<', $name)|| die "Unable to open file $name for locking: $!";
-    flock($lck, LOCK_SH);
-
     my $raw = IO::Uncompress::AnyUncompress->new($name)
       or die "Decompression failed: $AnyUncompressError\n";
 
@@ -251,24 +269,23 @@ sub pdiff {
 	$sha1->add($_);
     }
     close($raw);
-    flock($lck, LOCK_UN);
-    close($lck,);
+
     if ($AnyUncompressError) {
 	warn "$name Read failed: $AnyUncompressError. Aborting patch\n";
 	return;
     }
 	
     $digest = $sha1->hexdigest;
-#    printmsg "$basename SHA1: $digest\n";
+    # printmsg "$basename SHA1: $digest\n";
 
     # Read diff_Index
     my (@hist, @patch);
 
     open(my $diffin_fh, '<', $diffindex) || die("Cannot open $diffindex: $!");
-    flock($diffin_fh, LOCK_EX);
+    _flock($diffin_fh, LOCK_SH);
     my $diffindex_digest = $sha1->addfile($diffin_fh)->hexdigest;
     if ($diffindex_digest ne $diffindex_sha1) {
-	flock($diffin_fh, LOCK_UN);
+	_flock($diffin_fh, LOCK_UN);
 	close ($diffin_fh);
 	if ($force) {
 	    warn "$diffindex incorrect SHA1: expected $diffindex_sha1, got $diffindex_digest. Continuing anyway as --force specified\n";
@@ -278,14 +295,14 @@ sub pdiff {
 	return;
 	}
     }
-    seek($diffin_fh,0,0); # rewind
+    seek($diffin_fh,0,0) || die "Seek failed: $!"; # rewind
     my $curr= <$diffin_fh>; # read first line
     chomp $curr; # remove trailing \n
 #    printmsg "$diffindex: $curr\n";
     my ($target_sha1, $target_size) = (split (/\s+/,$curr))[1,2];
     if ($digest eq $target_sha1) { # check this matches /SHA1/
 	printmsg "SHA1 match: $name already up to date\n";
-	flock($diffin_fh, LOCK_UN);
+	_flock($diffin_fh, LOCK_UN);
 	close ($diffin_fh);
 	return 1; # success
     }
@@ -301,7 +318,7 @@ sub pdiff {
 	    next;
 	}
     }
-    flock($diffin_fh, LOCK_UN);
+    _flock($diffin_fh, LOCK_UN);
     close ($diffin_fh);
 
     my $diff;
@@ -310,7 +327,7 @@ sub pdiff {
 	my @line;
 	@line = split;
 #	printmsg "Checking $digest against @line\n";
-	if ( $digest eq $line[0]) {
+	if ($digest eq $line[0]) {
 #	    printmsg "found SHA1 match at \$hist $count: $line[0]\n";
 	    $diff = $count;
 	    last
@@ -380,7 +397,7 @@ sub pdiff {
 	warn "$name patching failed! $tfh is not size $target_size\n";
 	return;
     }
-    seek($tfh,0,0); # rewind
+    seek($tfh,0,0) || die "Seek failed: $!"; # rewind
     $sha1->addfile($tfh);
     $digest=$sha1->hexdigest;
     if ($digest eq $target_sha1) {
@@ -389,41 +406,34 @@ sub pdiff {
 	    printmsg "Simulation mode, so not replacing existing files\n";
 	}
 	else {
-	    my $destfile = "../temp/".$basename.".new".$type;
-	    printmsg "Saving as $destfile\n";
-	    printmsg "Warning: $destfile already exists\n" if -f $destfile;
-	    seek($tfh,0,0); # rewind
+	    printmsg "Saving as $name\n";
+	    seek($tfh, 0, 0) || die "Seek failed: $!"; # rewind
+	    truncate($cfh, 0) || die "Truncate failed: $!";
+	    seek($cfh, 0, 0) || die "Seek failed: $!";
 	    my ($z,$encoding) = ($name=~/bz2$/ ?
-				 ((IO::Compress::Bzip2->new($destfile)), "x-bzip2") :
+				 ((IO::Compress::Bzip2->new($cfh)), "x-bzip2") :
 				 ($name=~/gz$/ ?
-				  ((IO::Compress::Gzip->new($destfile, -Level => 9)), "x-gzip") :
-				  IO::Handle->new($destfile)));
+				  ((IO::Compress::Gzip->new($cfh, -Level => 9)), "x-gzip") :
+				  $cfh));
 
 	    while (<$tfh>) {
 		$z->print($_);
 	    }
 	    close($z);
-	    my $size = -s $destfile;
+	    $cfh->flush; # So the size is correct
 	    my $datestring = HTTP::Date::time2str;
 	    my $response = HTTP::Response->new(200, 'OK', ['Date' => $datestring,
-							   'Content-Length' => $size,
+							   'Content-Length' => -s $cfh,
 							   'Content-Type' => 'text/plain',
 							   'Last-Modified' => $datestring]);
 	    $response->header('Content-Encoding' => $encoding) if $encoding;;
-	    write_header("$destfile.header", $response);
-	    printmsg("Linking to $name\n");
-	    # Get global lock and copy
-	    set_global_lock('copy patched file');
-	    unlink $name,  "../headers/$name";
-	    link $destfile, $name || die "Link $destfile failed: $!";
-	    link "$destfile.header", "../headers/$name" || die "Link $destfile.header failed: $!";
-	    release_global_lock();
-	    printmsg "Unlink temporary files\n";
-	    unlink $destfile, "$destfile.header";
+	    write_header($hfh, $response);
+	    _flock($cfh, LOCK_SH); # Downgrade
+	    _flock($hfh, LOCK_UN);
 	    # Read checksums
 	    if ($cfg->{checksum}) {
 		printmsg ("Importing new checksums from patched $name\n");
-		import_sums($name);
+		import_sums($name, $cfh);
 	    }
 	}
     }
@@ -538,7 +548,7 @@ if (@db_mode || $db_recover){
 	    printmsg "Waiting for exclusive lock...";
 	    if (db_flock(LOCK_EX)){
 		printmsg "Got it!\nVerifying database...";
-		printmsg db_verify("$cfg->{cache_dir}/sums.db", temp_env()) ? "Failed! $!\n" : "Passed!\n";
+		printmsg db_verify("$cfg->{cache_dir}/sums.db", temp_env()) ? "Failed! " . db_error() . "\n" : "Passed!\n";
 	    }
 	    else {
 		warn "Unable to get exclusive database lock: $!\n";
@@ -561,7 +571,7 @@ if (@db_mode || $db_recover){
 
 open  (my $cleanup_lock, '<', $cfg->{'cache_dir'})
   or die "Can't open $cfg->{'cache_dir'} for locking!\nError: $!\n";
-flock $cleanup_lock, LOCK_EX|LOCK_NB
+flock($cleanup_lock, LOCK_EX|LOCK_NB)
   or die "Another apt-cacher-cleanup is already running. Exiting!\n";
 
 # check offline mode in config
@@ -571,14 +581,10 @@ if (defined $cfg->{offline_mode} && $cfg->{offline_mode}) {
 
 use GDBM_File;
 open(my $tmpfile, "+>", undef) or die $!;
-tie my %valid, 'GDBM_File', '/dev/fd/'.fileno($tmpfile), &GDBM_NEWDB|&GDBM_FAST, oct(600); # Does a separate open
+tie my %valid, 'GDBM_File', fd_path($tmpfile), &GDBM_NEWDB|&GDBM_FAST, oct(600) # Does a separate open
+  or die "GDBM_File tie failed: $!";
 close($tmpfile); # So we can close this
 
-my $tempdir="$cfg->{cache_dir}/temp";
-mkdir $tempdir if !-d $tempdir;
-die "Could not create tempdir $tempdir\n" if !-d $tempdir;
-unlink (glob("$tempdir/*"));
-
 ### Preparation of the package lists ########################################
 
 chdir "$cfg->{cache_dir}/packages" && -w "." || die "Could not enter the cache dir: $!";
@@ -611,10 +617,7 @@ if($> == 0 && !$cfg->{user} && !$force) {
     get($_) foreach keys %missing;
 }
 
-# file state decisions, lock that area
-set_global_lock('file state decision');
 my @ifiles = glob('*{Release,diff_Index} *{Packages,Sources}{,.gz,.bz2}');
-release_global_lock();
 
 foreach my $file (@ifiles) {
 
@@ -641,35 +644,13 @@ foreach my $file (@ifiles) {
 foreach my $file (@ifiles) {
     printmsg "Reading: $file\n";
 
-    # get both locks and create a temp. copy
-    my $tmpfile= "$tempdir/$file";
-    set_global_lock('temporary copy');
-    open(my $lck, '<', $file) || do {
-	release_global_lock();
-	print ("Error: cannot open $file for locking: $!\n");
+    (my ($cfh, $hfh) = open_lock($file)) == 2 || do {
+	die "Failed to open filehandles for $file. Delete this manually or use --force if the repository is no longer interesting. \nExiting to prevent deletion of cache contents.\n" unless $force;
+	warn "Forced ignoring empty index file $file, apparently undownloadable. All packages referenced by it will be lost!\n";
 	next;
     };
-    flock($lck, LOCK_EX);
-    link($file, $tmpfile) || do {
-	release_global_lock();
-	print ("Cannot link $file $tmpfile. Check permissions. $cfg->{cache_dir} must be single filesystem.\n");
-	next;
-    };
-    flock($lck, LOCK_UN);
-    close($lck);
-    release_global_lock();
 
-    if(-e $tmpfile && -z $tmpfile && $tmpfile=~/(?:gz|bz2)$/) {
-	# moo, junk, empty file, most likely leftovers from previous versions
-	# of apt-cacher-cleanup where the junk was "protected" from being
-	# deleted. Purge later by not having in %valid.
-	# delete $valid{$file}; <- will be recreated RSN either way
-      die("Found empty index file $file. Delete this manually or use --force if the repository is no longer interesting. \nExiting to prevent deletion of cache contents.\n") unless $force;
-      print "Forced ignoring empty index file $file, apparently undownloadable. All packages referenced by it will be lost!\n";
-    }
-    else {
-	extract_sums($tmpfile, \%valid) || die("Error processing $file in $cfg->{cache_dir}/packages, cleanup stopped.\nRemove the file if the repository is no longer interesting and the packages pulled from it are to be removed.\n");
-    }
+    extract_sums($file, $cfh, \%valid) || die("Error processing $file in $cfg->{cache_dir}/packages, cleanup stopped.\nRemove the file if the repository is no longer interesting and the packages pulled from it are to be removed.\n");
 }
 
 printmsg "Found ".scalar (keys %valid)." valid file entries\n";
@@ -709,9 +690,9 @@ foreach (glob('*{,/*}')) {
 	printmsg "Removing expired changelog: $_ and company...\n";
 	next;
     }
-    else {
-	next unless is_file_type('package', $_); # Package files only	
-    }
+
+    next unless is_file_type('package', $_) || is_file_type('pdiff', get_original_url($_)); # Package and pdiff files only
+
     if(! defined($valid{$_})) {
 	unlink $_, "../headers/$_", "../private/$_.complete" unless $sim_mode;
 	printmsg "Removing file: $_ and company...\n";
@@ -723,11 +704,15 @@ foreach (glob('*{,/*}')) {
 	# print "Validating SHA1 $target_sum for $_\n";
 	open(my $fh, '<', $_) || die "Unable to open file $_ to verify checksum: $!";
 	flock($fh, LOCK_EX);
+	if (is_file_type('pdiff', get_original_url($_))) { # pdiffs need decompressing
+	    $fh = IO::Uncompress::AnyUncompress->new($fh)
+	      or die "Decompression failed: $AnyUncompressError\n";
+	}
 	if ((my $sha1 = Digest::SHA->new(1)->addfile($fh)->hexdigest) ne $target_sum) {
 	    unlink $_, "../headers/$_", "../private/$_.complete" unless $sim_mode;
 	    printmsg "Checksum mismatch ($target_sum <=> $sha1): $_, removing\n";
 	}
-	flock($fh, LOCK_UN);
+	# No explicit LOCK_UN: it fails with IO::Uncompress::AnyUncompress, just rely on close
 	close $fh;
     }
 }
@@ -760,9 +745,8 @@ foreach (glob('*.complete')) {
       printmsg "Removing: $_.complete\n";
       unlink "$_.complete" unless $sim_mode;
    }
-   elsif (-e "../headers/$_") {
-       my $resp = read_header($_);
-       next if !$resp || !$resp->header('X-AptCacher-URL');
+   elsif (my $resp = read_header("../headers/$_")) {
+       next unless $resp->header('X-AptCacher-URL');
        # Remove complete files if we now have the upstream URL in the headers
        printmsg "Removing redundant $_.complete\n";
        unlink "$_.complete" unless $sim_mode;
@@ -781,16 +765,14 @@ foreach (glob('*{,/*}')) {
     # must be empty and not complete and being downloaded right now
     if(-z $_) {
 	my $fromfile;
-	if(open($fromfile, '<', $_) && flock($fromfile, LOCK_EX|LOCK_NB)) {
+	if(open($fromfile, '<', $_) && flock($fromfile, LOCK_SH|LOCK_NB)) {
 	    # double-check, may have changed while locking
-	    if(-z $_) {
+	    if(-z $fromfile) {
 		printmsg "Removing zombie files: $_ and company...\n";
 		unlink $_, "../headers/$_", "../private/$_.complete" unless $sim_mode;
-		flock($fromfile, LOCK_UN);
+		_flock($fromfile, LOCK_UN);
 		close($fromfile);
 	    }
 	}
     }
 }
-
-unlink (glob("$tempdir/*"));
