@@ -50,7 +50,7 @@ use warnings;
 
 # Include the library file
 use lib '/usr/share/apt-cacher/lib';
-require('apt-cacher-lib.pl');
+require('apt-cacher.pl');
 
 use Getopt::Long qw(:config no_ignore_case bundling);
 use File::Copy;
@@ -223,7 +223,7 @@ sub read_indices {
 	    print "Reading: $indexfile [namespace $namespace]\n" unless $quiet;
 
 	    # Parse
-	    extract_sums($indexfile, \%sums)
+	    extract_sums($indexfile, undef, \%sums)
 	      || die("Error processing $indexfile, automatic digest import abandoned.\nRemove the file if the packages to be imported are not associated with this repository.\n");
 
 	    while (my ($package, $ice) = each %sums) {
