@@ -37,7 +37,7 @@ my %options = (
 &help if ($help);
 
 # Include the library for the config file parser
-require '/usr/share/apt-cacher/apt-cacher-lib.pl';
+require '/usr/share/apt-cacher/lib/apt-cacher.pl';
 my $cfgref;
 eval {
         $cfgref = read_config($configfile);
