@@ -7,8 +7,8 @@
 
 use strict;
 use warnings;
-use POSIX;
-use Fcntl qw/:flock/;
+use POSIX ();
+use Fcntl qw/:DEFAULT :flock/;
 use FreezeThaw qw(freeze thaw);
 use HTTP::Response;
 use URI;
@@ -42,7 +42,7 @@ sub read_config {
 		  denied_hosts => '',
 		  distinct_namespaces => 0,
 		  expire_hours => 0,
-		  fetch_timeout => 300, # five minutes from now
+		  data_timeout => 120,
 		  generate_reports => 1,
 		  group => eval {my $g = $); $g =~ s/\s.*$//; $g},
 		  http_proxy => '',
@@ -157,11 +157,13 @@ sub read_config {
 	close $fh;
     }
 
-    if ($config{logdir}) { # Recognise the old name
-	$config{log_dir} = $config{logdir};
-	delete $config{logdir};
+    # Recognise old/renamed configuration options
+    foreach (['logdir' => 'log_dir'], ['fetch_timeout' => 'data_timeout']) {
+        if ($config{@$_[0]}) {
+	    $config{@$_[1]} = $config{@$_[0]};
+	    delete $config{@$_[0]};
+	}
     }
-
     return \%config;
 }
 
@@ -187,6 +189,17 @@ sub private_config {
 	}
     }
 
+    # Handle libcurl configuration
+    if ($cfg->{libcurl}) {
+        for (cfg_split($cfg->{libcurl})) {
+	    my @tmp = split(/\s+/, $_);
+	    next unless my $key=uc(shift(@tmp));
+	    if (@tmp) {
+		$cfg->{_libcurl}{$key} = shift(@tmp);
+	    }
+	}
+    }
+
     # Expand PATH_MAP in allowed_hosts
     if ($cfg->{allowed_locations}) {
         $cfg->{allowed_locations} =~ s/\bPATH_MAP\b/join(', ', keys %{$cfg->{_path_map}})/ge;
@@ -274,8 +287,8 @@ sub check_install {
     # Die if we have not been configured correctly
     die "$0: No cache_dir directory!\n" if (!-d $cfg->{cache_dir});
 
-    my $uid = $cfg->{user}=~/^\d+$/ ? $cfg->{user} : getpwnam($cfg->{user});
-    my $gid = $cfg->{group}=~/^\d+$/ ? $cfg->{group} : getgrnam($cfg->{group});
+    my $uid = $cfg->{user}=~/^\d+$/ ? $cfg->{user} : POSIX::getpwnam($cfg->{user});
+    my $gid = $cfg->{group}=~/^\d+$/ ? $cfg->{group} : POSIX::getgrnam($cfg->{group});
 
     if (!defined ($uid || $gid)) {
 	die "Unable to get user:group";
@@ -283,7 +296,7 @@ sub check_install {
 
     my @dir = ($cfg->{cache_dir}, $cfg->{log_dir}, "$cfg->{cache_dir}/private",
 		     "$cfg->{cache_dir}/import", "$cfg->{cache_dir}/packages",
-		     "$cfg->{cache_dir}/headers", "$cfg->{cache_dir}/temp");
+		     "$cfg->{cache_dir}/headers");
     foreach my $dir (@dir) {
 	if (!-d $dir) {
 	    print "Info: $dir missing. Doing mkdir($dir, 0755)\n";
@@ -323,25 +336,102 @@ sub hashify {
     }
 }
 
-# returns HTTP::Response
+# Get filename from filehandle
+sub filename_fh {
+    my ($fh) = @_;
+
+    return readlink fd_path($fh);
+}
+
+# Get path of fildescriptor
+# VERY Linux specific
+sub fd_path {
+    my ($fh) = @_;
+
+    die 'Not a GLOB' unless ref $fh eq 'GLOB';
+    return '/dev/fd/' . $fh->fileno;
+}
+
+# Delete cached files by filehandle
+sub unlink_by_fh {
+    my @fh = @_;
+
+    my $count;
+
+    foreach (@fh) {
+	next unless ref eq 'GLOB';
+	$_ =  filename_fh($_);
+	debug_message("Deleting $_") if defined &debug_message;
+	($count += unlink $_) || warn "Failed to delete $_: $!";
+    }
+    return $count;
+}
+
+
+# Verbose wrapper to flock
+sub _flock {
+    my ($fh, $flags) = @_;
+
+    my $ret;
+    unless ($ret = flock($fh, $flags | LOCK_NB)) {
+	if ($cfg->{debug}) {
+	    # Hash for decoding flag symbols.
+	    #
+	    # __PACKAGE__->$sym references flock constants without needing to
+	    # disable strict refs
+	    my %h = map {$_ => __PACKAGE__->$_} glob("LOCK_{SH,EX,UN,NB}");
+	    debug_message('Waiting for '
+			  . join ('|', grep {$h{$_} & $flags && $_} keys %h)
+			  . ' on '
+			  . filename_fh($fh)) if defined &debug_message;
+	}
+        $ret = flock($fh, $flags);
+	debug_message('Got it!') if defined &debug_message;
+    }
+    return $ret;
+}
+
+# Argument is filehandle or filename, returns HTTP::Response
 sub read_header {
     my ($file) = @_;
-    my $r;
-    if ($file && open(my $fh, '<', $file)) {
-	local $/; # Slurp
-	$r = HTTP::Response->parse(<$fh>);
-	close($fh);
-	chomp_message($r);
-	# Fake Client-Date if not specified
-	$r->client_date($r->date || time) unless $r->client_date;
-	$r->header('Age' => $r->current_age);
+    my ($r, $fh);
+
+    if (ref $file eq 'GLOB') {
+	$fh = $file;
     }
+    else {
+	open($fh, '<', $file) || die "Open header $file failed: $!";
+    }
+
+    if ($fh) {
+	for ($fh->getline) {
+	    last unless defined;
+	    if (/^(HTTP\/1\.[01]\s+)?\d{3}\s+/) { # Valid
+		seek($fh,0,0) || die "Seek failed: $!";
+		{
+		    local $/; # Slurp
+		    $r = HTTP::Response->parse(<$fh>);
+		}
+		chomp_message($r);
+		# Fake Client-Date if not specified
+		$r->client_date($r->date || time) unless $r->client_date;
+		$r->header('Age' => $r->current_age);
+	    }
+	    else { # Invalid
+		warn "Invalid/corrupt header file: $file";
+		undef $r;
+		undef $fh;
+	    }
+	}
+    }
+
+    # Don't explicitly close $fh
     return $r;
 }
 
-# Args are filename and HTTP::Response
+# Args are filename/filehandle and HTTP::Response
 sub write_header {
-    my ($filename, $response) = @_;
+    my ($file, $response) = @_;
 
     debug_message('Writing header') if defined (&debug_message);
     # Remove Connection header and options
@@ -355,10 +445,18 @@ sub write_header {
 	$response->header('X-AptCacher-URL' => $request_url);
     }
     $response->client_date(time); # HTTP::Response uses this to calculate current_age
-    open (my $chfh, '>', $filename) || barf("Unable to open $filename, $!");
+    my $chfh;
+    if (ref $file eq 'GLOB'){
+	$chfh = $file;
+	$chfh->truncate(0) || die "Truncate failed: $!";
+	seek($chfh,0,0) || die "Seek failed: $!";
+    }
+    else {
+	open ($chfh, '>', $file) || die "Unable to open $file, $!";
+    }
     print $chfh $response->status_line, "\n";
     print $chfh $response->headers->as_string;
-    close($chfh);
+    # No explicit close. Rely on gc or explicit close in caller
     return;
 }
 
@@ -394,8 +492,8 @@ sub get_namespace {
     return;
 }
 
-# Returns URI object of url used to fetch file
-sub get_original_url {
+# Returns URI object of url used by libcurl to fetch file
+sub get_upstream_url {
     my ($filename) = @_;
     my $uri;
 
@@ -403,30 +501,45 @@ sub get_original_url {
     if (my $response = read_header("$cfg->{cache_dir}/headers/$filename")) {
 	$uri = URI->new($response->header('X-AptCacher-URL'));
     }
-    # Old complete file
-    if (!$uri && -f (my $complete_file = "$cfg->{cache_dir}/private/$filename.complete")) {
-	open(my $cfh, '<', $complete_file) || die "Failed to open $complete_file:$!";
-	$uri = URI->new(<$cfh>);
-	close($cfh);
-    }
-    else {
-	# Infer from filename, assume HTTP
-	$uri = URI->new('http://' . join('/', split(/_/, $filename)));
+
+    unless ($uri) {
+	# Old complete file
+	if (open (my $cfh, '<', my $complete_file = "$cfg->{cache_dir}/private/$filename.complete")) {
+	    $uri = URI->new(<$cfh>);
+	    close($cfh);
+	}
+	else {
+	    # Assume same as request
+	    $uri = get_original_url($filename);
+	}
     }
     return $uri;
 }
 
-# Stores data flattened for use in tied hashes
-sub extract_sums {
-   my ($name, $hashref) = @_;
+# Returns URI object of url used to request file
+sub get_original_url {
+    my ($filename) = @_;
 
-   seek($name,0,0) if fileno($name);
-   my $raw = IO::Uncompress::AnyUncompress->new($name)
+    # Infer from filename, assume HTTP
+    return URI->new('http://' . join('/', split(/_/, $filename)));
+}
+
+# Stores data flattened for use in tied hashes
+# Arg $fh is optional
+sub extract_sums {
+   my ($name, $fh, $hashref) = @_;
+
+   if ($fh) {
+       seek($fh,0,0) || die "Seek failed: $!";
+   }
+   else {
+       open($fh, '<', $name) || die "Open $name failed: $!";
+   }
+
+   my $raw = IO::Uncompress::AnyUncompress->new($fh)
      or die "Decompression failed: $AnyUncompressError\n";
 
-   # If arg is fd get full path
-   $name = readlink('/dev/fd/'.fileno($name)) if fileno($name);
-   # Name is just the cached filename
+   # Name is just the cached filename without path
    $name = (File::Spec->splitpath($name))[2];
 
    # Determine namespace
@@ -528,12 +641,12 @@ sub extract_sums {
 
 	debug_message("Global lock: \u$msg") if defined (&debug_message);
 
-	#may need to create it if the file got lost
-	open($glock, ((-f $glockfile) ? '<' : '>'), $glockfile) || die "Unable to open lockfile: $!";
-	if (!flock($glock, LOCK_EX)) {
-	    debug_message("Unable to lock $glockfile for \u$msg: $!") if defined (&debug_message);
-	    die "Unable to lock $glockfile for \u$msg: $!";
-	}
+	# May need to create it if the file got lost
+	sysopen($glock, $glockfile, O_CREAT|O_WRONLY) || die "Unable to open lockfile: $!";
+	$glock->autoflush;
+	_flock($glock, LOCK_EX) || die "Unable to lock $glockfile for \u$msg: $!";
+	$glock->truncate(0);
+	print $glock "$$: \u$msg\n";
 	return defined($glock);
     }
 
@@ -542,11 +655,15 @@ sub extract_sums {
 	    carp('Attmept to free lock not held');
 	    return;
 	}
-	flock($glock, LOCK_UN) || die "Unable to release lock: $!";
+	_flock($glock, LOCK_UN) || die "Unable to release lock: $!";
 	close $glock || die "Unable to close lock: $!";
 	debug_message("Release global lock") if defined (&debug_message);
 	return;
     }
+
+    sub global_lock_fh {
+	return $glock;
+    }
 }
 
 
@@ -567,37 +684,32 @@ sub setup_ownership {
 
     if($gid) {
 	if($gid=~/^\d+$/) {
-	    my $name=getgrgid($gid);
+	    my $name = POSIX::getgrgid($gid);
 	    die "Unknown group ID: $gid (exiting)\n" if !$name;
 	}
 	else {
-	    $gid=getgrnam($gid);
+	    $gid = POSIX::getgrnam($gid);
 	    die "No such group (exiting)\n" if !defined($gid);
 	}
-	setgid($gid) || barf("setgid failed: $!");;
-	$) =~ /^$gid\b/ && $( =~ /^$gid\b/ || barf("Unable to change group id");
+	POSIX::setgid($gid) || die "setgid failed: $!";
+	$) =~ /^$gid\b/ && $( =~ /^$gid\b/ || die "Unable to change group id";
     }
 
     if($uid) {
 	if($uid=~/^\d+$/) {
-	    my $name=getpwuid($uid);
+	    my $name = POSIX::getpwuid($uid);
 	    die "Unknown user ID: $uid (exiting)\n" if !$name;
 	}
 	else {
-	    $uid=getpwnam($uid);
+	    $uid = POSIX::getpwnam($uid);
 	    die "No such user (exiting)\n" if !defined($uid);
 	}
-	setuid($uid) || barf("setuid failed: $!");
-	$> == $uid && $< == $uid || barf("Unable to change user id");
+	POSIX::setuid($uid) || die "setuid failed: $!";
+	$> == $uid && $< == $uid || die "Unable to change user id";
     }
     return;
 }
 
-sub barf {
-	my $errs = shift;
-	die "--- $0: Fatal: $errs\n";
-}
-
 # Still matches against the filename only if called with a fully qualified path
 sub is_file_type {
     my ($type,$file) = @_;
